/*
 * Copyright (c) 1996-2007, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */


#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "headers.h"
#include "http.h"
#include "icap.h"
#include "url.h"
#include "base64.h"

#ifndef MAX
#define MAX(a, b) (((a) < (b)) ? (b) : (a))
#endif

#ifndef MIN
#define MIN(a, b) (((a) > (b)) ? (b) : (a))
#endif

#define SELECT_CTYPE_HTTP
#include "tables.h"

#include "ci_milter.h"
#include "ci_milter_sys.h"

#include "filters.h"

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

extern const unsigned char http_char_type[UCHAR_MAX+1];

static HDR*
new_HDR_from_hdr_line(struct hdr_line *hl)
{
	HDR	*h;
	char	*line;
	char 	*s;

	if (!hl)
		return(NULL);

	line = xstrdup(hl->line);
	s = index (line, ':');

	if (!s)
		return(NULL);

	h = (HDR *)xmalloc(sizeof(HDR));
	*(s++) = '\0';
	h->h_field = xstrdup(line);

	while (WHITE (*s))
		++s;
	h->h_value = xstrdup(s);
	return(h);

}

static int check_code_200(struct hdr_line *h)
{
	char *c = h->line;

/*	It is something like "HTTP/1.0 200 OK". So we skip to the
 *	first whitespace, then skip the whitespace and get number from there.
 */
	while (WHITE(*c)  && (c < (h->line + h->length )))
		c++;
	while (!WHITE(*c) && (c < (h->line + h->length)))
		c++;
	while (WHITE(*c) && (c < (h->line + h->length)))
		c++;

	if (!*c)
		return(-1);
	if (atoi(c) != 200)
		return(1);
	return(0);
	
}

/* 
 * Generic content inspection API, supports icap and milter engines.
 * bodyfile is mandatory at the moment, "headers only" requests are
 * not to be handled here. objsize is "write-only", the value is
 * ignored. http_request should contain something quite closely
 * resembling valid http request, containing method, scheme, uri and
 * protocol with version. May be overwriten, so watch out for buffer overflow,
 * it should be at least MAX_STR size. icap_req should be REQMOD or RESPMOD.
 * filters may be protocol:path for milters or icap:// uri, auto-detected.
 */ 
 

int filter_inspect_object(char* bodyfile,
		   int64_t* objsize,  char *http_request, 
		   icap_req icap_rq, char **filters, 
		   struct app_data *adata,
		   char** err_msg)
{
	char** flt;
	struct	hdr_line	*mockup_rq_header = NULL;	
	struct	hdr_line	*mockup_rs_header = NULL;	
	struct	hdr_line	*milter_header = NULL;	
	struct	hdr_line	*icap_header	  = NULL;
	char	buffer[MAX_STR + 1];
	int	previewsize 			  = 4096;

	/* We do some unneeded work if our filters are all non-icap.
	 * actually milter interfaces seem to become obsolete soon,
	 * so i doubt it is worth fixing.
	 */

	header_add_line(&mockup_rq_header, http_request);
	if ((icap_rq == RESPMOD) || !adata || !(adata->headers)) {
		snprintf(buffer, sizeof(buffer) - 1, "Host: %s", 
		         (adata && adata->target_host) ? 
			 	adata->target_host : proxy_stats.dst);
		header_add_line(&mockup_rq_header, buffer);
	}

	if (icap_rq == RESPMOD) {
		/* 
		 * We mock up some http-like response using protocol name
		 * and version from the request. Code 200 may be misleading,
		 * though, but i have no better idea that doesn not make 
		 * parameters complicated without any real purpose. Most
		 * likely, icap filters will just ignore that stuff. 
		 */
		char	*proto = NULL;
		if (!(proto = strrchr(http_request,' ')) || !(*(++proto)))
			proto = "HTTP/1.0";
		snprintf(buffer, sizeof(buffer) - 1, "%s 200 OK", proto);
		header_add_line(&mockup_rs_header, buffer);
		if (adata && adata->headers)
			milter_header = mockup_rs_header->next =adata->headers;
		else
			header_add_line(&mockup_rs_header,"Content-Type: application/octet-stream");
	} else if (icap_rq == REQMOD) {
		if (adata && adata->headers)
			milter_header = mockup_rq_header->next =adata->headers;
	} else {
		syslog(LLEV,"fwtksyserr: internal logic error: invalid icap request for mail inspection");
		*err_msg = "Internal error";
		return(-1);
	}

	for (flt = filters; *flt && **flt; flt++) {
		struct url u;
		static int modified = 0;
		if (url_parse(&u, (octet *) *flt, strlen(*flt), 0) ||
		    (u.scheme != URL_ICAP)) {
			int rc;
			HDR* h = NULL;
			HDR* h1 = NULL;
			struct hdr_line *mh;
			/* milter */
			for (mh = milter_header; mh ; mh = mh->next) {
				if(!h1) {
					h = new_HDR_from_hdr_line(mh);
					h1 = h;
				} else {
					h1->h_link = new_HDR_from_hdr_line(mh);
					h1 = h1->h_link;
				}
			}
				    			
			rc = milter_inspect(proxy_name, proxy_stats.rladdr,
					    proxy_stats.riaddr, &h, bodyfile,
					    adata->encode,
					    adata->smtp_sender, 
					    &(adata->smtp_rcpts), *flt);
			if (rc & MS_ER) {
				syslog(LLEV,"fwtksyserr: milter %s failed on %s", *flt, bodyfile);
	  
				if(err_msg)
					*err_msg = "Content inspection failed";

				return(-1);
			}
			if (rc & MS_QU) {
				syslog(LLEV,"milter %s requested quarantine of %s", *flt, bodyfile);
	  	
				if(err_msg)
					*err_msg = "Content quarantined by filter request";
	
				return(1); 
			}
			if (rc & MS_RJ) {
				if(err_msg)
					*err_msg = "Content rejected";
			
				return(1);
			}
			if (rc & OK_HC) {
				header_free(&(adata->headers));
				for (h1 = h; h1; h1 = h1->h_link) {
					snprintf(buffer, MAX_STR, "%s: %s",
						 h1->h_field, h1->h_value);
					header_add_line(&(adata->headers), buffer);
				}
			}
			/* No special action required for (rc & OK_BC)
			 * the message body is silently rebuilt if possible
			 * encoded body modification is not supported.
			 */
			if ((rc & OK_BC)  && adata->encode) {
				syslog(LLEV,"fwtksyserr: internal logic error -- invalid request from milter API (modification of encoded message)");
				*err_msg = "Internal error";
				return(-1);
			}
		} else {
			int 			rc;
			EMI_FILE		*icap_in;
			EMO_FILE		*icap_out;
			EMI_FILE		*body;
			EMO_FILE		*body_w;

			if (icap_open_connection(*flt, &icap_in, &icap_out)) {
				syslog(LLEV,"fwtksyserr: cannot connect to icap server");
				*err_msg = "Content inspection failed";
				return(-1);
			}

			if ((rc = icap_request(icap_in, icap_out, *flt,
					       OPTIONS, NULL, NULL, NULL, 0, 0,
					       NULL, 0, &icap_header)) != 200) {
			} else {
				char *pv = header_field_value(header_field_by_name(icap_header, "Preview:"), NULL);
				if (pv)
					previewsize = atoi(pv);
			}

			if (bodyfile)
				body = emi_open(bodyfile, MAX(previewsize,4096));

			switch (rc = icap_request(icap_in, icap_out, *flt, 
						  icap_rq, mockup_rq_header,
						  mockup_rs_header, 
						  bodyfile ? body : NULL, 0,
						  -1, &bodyfile,
						  previewsize, &icap_header)) {
			case 204:
				if (bodyfile)
					emi_close(body);
				break;
			case 200:
				if (bodyfile)
					emi_close(body);
				modified = 1;
				if ((icap_rq == REQMOD) && 
				    icap_is_response(icap_header)) {
				/* 
				 * This is pretty generic interface,
				 * so we do not hadle fancy error pages from 
				 * icap. Not here, or at least not yet,
				 * use raw icap_request api
				 * if you need to. 
				 */
				    	*err_msg = "Request rejected";
					emi_close(icap_in);
					emo_close(icap_out);
					return(1);
				}

				/* Skip to body (and ensure it exists)	*/
				if (bodyfile)
					body_w = emo_open(bodyfile, 4096);
				if (icap_rq == REQMOD) {
					if (mockup_rq_header)
						header_free(&mockup_rq_header);	
					mockup_rq_header = header_read(icap_in,"icap");
					strlcpy(http_request, 
						mockup_rq_header->line,
						MIN(MAX_STR-1, 
						    mockup_rq_header->length));
					if (!mockup_rq_header || 
					    !(mockup_rq_header->next)) {
					} else if (adata)
						adata->headers = mockup_rq_header->next;
				} else {
					if (mockup_rs_header)
						header_free(&mockup_rs_header);	
					mockup_rs_header = header_read(icap_in,"icap");
					if (!mockup_rs_header || 
					    !(mockup_rs_header->next)) {
					    	syslog(LLEV,"fwtksyserr: cannot read icap response header");
					    	*err_msg = "Content inspection falied";
						emi_close(icap_in);
						emo_close(icap_out);
						return(-1);
					} else if (adata)
						adata->headers = mockup_rs_header->next;
					if (check_code_200(mockup_rs_header)) {
				    		*err_msg = "Content rejected";
						emi_close(icap_in);
						emo_close(icap_out);
						return(1);
					}
				}
				if (bodyfile) {
					http_copy(body_w, 0, icap_in, 1, 
					  	MAX_CONTENT_LENGTH, NONE);
					if (objsize)
						*objsize = emo_amount(body_w,0);
					emo_close(body_w);
				}
				break;
			case -1:
				/* Error */
				syslog(LLEV,"fwtksyserr: icap failed");
			default:
				if (bodyfile)
					emi_close(body);
				syslog(LLEV,"icap returned code %d", rc);
				emi_close(icap_in);
				emo_close(icap_out);
				*err_msg = "Content inspection failed";
				return(-1);
			}
			/* Contents are already replaced by ICAP     	*/
			/* (if necessary)				*/
			emi_close(icap_in);
			emo_close(icap_out);
		}
	}
	return 0;
}
		
int inspect_file (char* file2inspect, size_t *fsize,
		  char *http_request, char *target_host,
		  icap_req icap_rq,
	      	  char* ctype, char** filters, char** err_msg)
{
	char** flt;
	struct	hdr_line	*mockup_rq_header = NULL;	
	struct	hdr_line	*mockup_rs_header = NULL;	
	struct	hdr_line	*icap_header	  = NULL;
	char	buffer[MAX_STR + 1];
	int	previewsize = 4096;
	size_t	bodysize 	= (size_t) -1;
	
	char **rcpt = xmalloc(sizeof(char*) * 2);
	snprintf(buffer, MAX_STR, "%s@%s", proxy_stats.authuser, "gateway");
	rcpt[0] = xstrdup(buffer);
	rcpt[1] = NULL;
      	
	header_add_line(&mockup_rq_header, http_request);
	snprintf(buffer, sizeof(buffer) - 1, "Host: %s", target_host);
	header_add_line(&mockup_rq_header, buffer);

	if (icap_rq == RESPMOD) {
		/* 
		 * We mock up some http-like response using protocol name
		 * and version from the request. Code 200 may be misleading,
		 * though, but i have no better idea that doesn not make 
		 * parameters complicated without any real purpose. Most
		 * likely, icap filters will just ignore that stuff. 
		 */
		char	*proto = NULL;
		if (!(proto = strrchr(http_request,' ')) || !(*(++proto)))
			proto = "HTTP/1.1";
		snprintf(buffer, sizeof(buffer) - 1, "%s 200 OK", proto);
		header_add_line(&mockup_rs_header, buffer);
		snprintf(buffer, sizeof(buffer) - 1, "Content-Type: %s", ctype ? ctype :
								"application/octet-stream");
		header_add_line(&mockup_rs_header, buffer);
		if (fsize && *fsize && (*fsize != -1)) {
			snprintf(buffer, sizeof(buffer) - 1, "Content-Length: %u", (unsigned int) *fsize);
			header_add_line(&mockup_rs_header, buffer);
			bodysize = *fsize;
		}
	} else {
		snprintf(buffer, sizeof(buffer) - 1, "Content-Type: %s", ctype ? ctype :
								"application/octet-stream");
		header_add_line(&mockup_rq_header, buffer);
		if (fsize && *fsize && (*fsize != -1)) {
			snprintf(buffer, sizeof(buffer) - 1, "Content-Length: %u", (unsigned int) *fsize);
			header_add_line(&mockup_rq_header, buffer);
		}
	}

	for (flt = filters; *flt; flt++) {
		struct url u;
		if (url_parse(&u, (octet *) *flt, strlen(*flt), 0) ||
		    (u.scheme != URL_ICAP)) {
			/* milter */
			int rc;	
	
			rc = milter_inspect_binary(proxy_name, proxy_stats.rladdr,
					proxy_stats.riaddr,  *flt, file2inspect, "unauth@gateway",
					&rcpt, "attachment.dat", ctype ? ctype : "application/octet-stream");
		
			if (rc & MS_ER) {
				syslog(LLEV,"fwtksyserr: milter %s failed on %s", *flt, file2inspect);
	  
				if(err_msg)
					*err_msg = "Content inspection failed";

				return -1;
			}
			if (rc & MS_QU) {
				syslog(LLEV,"milter %s requested quarantine of %s", *flt, file2inspect);
	  	
				if(err_msg)
					*err_msg = "Content quarantined by filter request";
	
				return 1; 
			}
			if (rc & MS_RJ) {
				if(err_msg)
					*err_msg = "Content rejected";
			
				return 1;
			}
		} else {
			int 			rc;
			EMI_FILE		*icap_in;
			EMO_FILE		*icap_out;
			EMI_FILE		*body;
			EMO_FILE		*body_w;

			if (icap_open_connection(*flt, &icap_in, &icap_out)) {
				syslog(LLEV,"fwtksyserr: cannot connect to icap server");
				*err_msg = "Content inspection failed";
			}

			if ((rc = icap_request(icap_in, icap_out, *flt,
					       OPTIONS, NULL, NULL, NULL, 0, 0,
					       NULL, 0, &icap_header)) != 200) {
			} else {
				char *pv = header_field_value(header_field_by_name(icap_header, "Preview:"), NULL);
				if (pv)
					previewsize = atoi(pv);
			}

			body = emi_open(file2inspect, MAX(previewsize,4096));

			switch (rc = icap_request(icap_in, icap_out, *flt, 
						  icap_rq, mockup_rq_header,
						  mockup_rs_header, body, 0,
						  bodysize, &file2inspect,
						  previewsize, &icap_header)) {
			case 204:
				emi_close(body); 
				/* No modification required		    */
				/* Reopen (emi_* does not include seek api) */
				body = emi_open(file2inspect, previewsize);
				break;
			case 200:
				emi_close(body);
				if ((icap_rq == REQMOD) && 
				    icap_is_response(icap_header)) {
				/* Error				*/
				}
				/* Skip to body (and ensure it exists)	*/
				if (icap_rq == REQMOD) {
					if (mockup_rq_header)
						header_free(&mockup_rq_header);	
					mockup_rq_header = header_read(icap_in,"icap");
				} else {
					if (mockup_rs_header)
						header_free(&mockup_rs_header);	
					mockup_rs_header = header_read(icap_in,"icap");
				}
				body_w = emo_open(file2inspect, 4096);
				http_copy(body_w, 0, icap_in, 1, 
					  MAX_CONTENT_LENGTH, NONE);
				if (fsize)
					*fsize = emo_amount(body_w, 0);
				emo_close(body_w);
				body = emi_open(file2inspect, previewsize);
				break;
			case -1:
				/* Error */
				syslog(LLEV,"fwtksyserr: icap failed");
			default:
				emi_close(body);
				syslog(LLEV,"icap returned code %d", rc);
				emi_close(icap_in);
				emo_close(icap_out);
				return(rc);
			}
			emi_close(icap_in);
			emo_close(icap_out);
		}
	}
	
	return 0;
}

