/*
 * Copyright (c) 2009-2010 Security Technology Research
 * All rights reserved.
 */
/*
 * Copyright (c) 1997-2010, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/* Generic rfc2822/rfc2616 header support functions */

/* Copyright 1997-2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"
#include "auth.h"
#include "ctype.h"
#include "headers.h"
#include "http.h"
#include "icap.h"
#include "message.h"

/* Statistics */

/* =========================== PRELIMINARY STUFF =========================== */

/* Maximum total size of header. */

#define MAX_HDR_SIZE		(128 * 1024)

/* Flag bits for method_flags in struct request_state. */

#define METHOD_REQUEST_BODY	0x0001 /* Request has a body */
#define METHOD_RESPONSE_BODY	0x0002 /* Response may have a body */
#define METHOD_CONNECT		0x0004 /* CONNECT, handled specially */


/* ============================== CHARACTERS =============================== */

/* Character classes as defined by RFC 1945. */

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

/* Also referenced by cookies.c. */

extern const unsigned char http_char_type[UCHAR_MAX+1];

/* =========================== INPUT AND OUTPUT ============================ */

char*
header_field_value (const struct hdr_line *hl, size_t *plen)
{
  	char 	*s;
	if (!hl)
		return(NULL);
 	s = memchr (hl->line, ':', hl->length);
  	DEBUG_ASSERT (s != NULL);
  	++s;
  	while (WHITE (*s))
    		++s;
  	if (plen != NULL)
    		*plen = (hl->line + hl->length) - s;
  	return(s);
}

struct hdr_line*
header_field_by_name (struct hdr_line *hl, const char *name)
{
	struct	hdr_line	*p;
	for (p = hl; p ; p = p->next) {
		if (!strncasecmp(name, p->line, strlen(name)))
			return (p);
	}
	return(NULL);
}


/* Read one line terminated by CR*LF (i.e., a LF character preceded by
   any number of CR characters) from F, storing it to the array of
   SIZE characters pointed to by BUF.  Log the line if MSG is not
   NULL.  The resulting string may contain embedded control characters
   (except for the null character, which is rejected).  The
   terminating CR*LF is removed.  This function returns the length of
   the line. */

static size_t 
read_line (EMI_FILE *f, char *buf, size_t size, const char *msg,
			 int lineno)
{
  	size_t 	i = 0;
  	int 	c;

  	while ((c = emi_getc (f)) != EMI_EOF && c != LF) {
      		if (i + 1 >= size) {
			syslog(LLEV,"%s header line too long", msg ? msg : "");
			return(-1);
		}
	/*
	 * check for unusual headers.
	 * these form the basis of a number of more interesting attacks
	 *    - Julian Assange <proff@suburbia.net>
	 */
      		if (!isalpha(c) && !isprint(c) &&
		    !isspace(c) && ((c & 0x7f) < 0x20)) {
			syslog(LLEV,"%s header line contains invalid character", msg);
			return(-1);
		}
      		buf[i++] = (char)c;
    	}
  	if (c == LF || (c == EMI_EOF && !emi_error (f) /* FIN */
		  && i == 0)) { /* empty line */
      /* End of line.  There are some pathological cases:

         The OSU server terminates HTTP header lines with CR CR LF.
         Ignore any number of CRs.
       */

      		if (c == EMI_EOF && msg)
			syslog (LLEV, "Missing blank line after %s header", msg);
      		while (i != 0 && buf[i-1] == CR)
        		--i;
      		buf[i++] = CR;
      		buf[i++] = LF;
      		buf[i] = 0;
      		return i;
    	}
  	if (!emi_error (f))
		syslog(LLEV, "premature end of %s header (line=%d length=%u bytes=%llu)",
	   		msg, lineno, (unsigned)i, emi_amount (f, 0)); 
    	return(-1);
}

/* Read multiple lines from F and return them as a linked list of
   lines.  Concatenate lines with their continuation lines.  Stop
   reading at the first empty line or a line consisting only dot ('.')
   character. Keep the dot itself at the end of linked list.
   This function returns a pointer to the first element of the linked list. 
   If msg is not NULL, treat it as header name for syslog */

struct hdr_line *
header_read (EMI_FILE *f, const char *msg)
{
  	char 		buffer[MAX_HDR_SIZE];
  	struct hdr_line *first, *p, **add;
  	size_t 		line_len, cur_len;
  	int 		lineno = 0;
  	int 		header_size = 0;

  	first = NULL; add = &first; 

  	/* Read the first line.  RFC 2068 (HTTP/1.1) says that empty leading
     	   lines should be ignored. */

  	do {
      		line_len = read_line (f, buffer, sizeof (buffer), msg, ++lineno);
    	} while (line_len == 0);

  	if (WHITE (buffer[0])) {
    		syslog(LLEV, "%s header starts with continuation line", msg);
		return(NULL);
	}


  	/* Read lines until an empty line is reached.  Lines starting with
           whitespace are continuation lines and are appended to the
           previous line. */

  	for (;;) {
		int fold = 0;
        /* In fact, we always append the next line to the current line.
           If it's a continuation line, this is exactly what is desired.
           If it's not a continuation line, flush the previous line from
           the buffer and start a new line.  The buffer already contains
           LINE_LEN characters at this point. */

      		DEBUG_ASSERT (line_len < sizeof (buffer));
      		cur_len = read_line (f, buffer + line_len, sizeof (buffer) - line_len,
			   msg, ++lineno);
		if (cur_len == -1)
			return(NULL);
		while (cur_len > MAXLINE) {
			char	*s, *last, *head, *tail;
			size_t i;
			fold = 1;
		/*	
		 *	Doing our best to fold a long line.
		 *	Well, maybe not really "best", just fair. We do not
		 *	care about things that may break: non-breakable
		 *	space in quotes and other stuff alike.
		 *	I doubt you ever encounter anything useful
		 *	beyond MAXLINE limit (well-behaving software is
		 *	expected to follow rc822 and do the folding itself)
		 *	though beware, here be dragons.
		 */
			s = buffer + line_len;
			for (last = s + MAXLINE; 
			     (last > s + 3) && !WHITE(*last) ; last--)
				;
			if (last <= s+3) {
		/*	
		 *	Fold failed
		 */
				syslog(LLEV,"%s header line too long", msg ? msg : "");
				return(NULL);
			}
			i = last - s;
		/* 
	   	 *	Truncate, split and fold	
		 */
			head = xstrdup(s);
			tail = head + i + 1;
			head[i] = '\0';
			buffer[line_len] = '\0';
			line_len += strlen(head) + 2;
			cur_len = strlen(tail);
			if (line_len + cur_len >= MAX_HDR_SIZE) {
		/*	
		 *	Fold failed
		 */
				syslog(LLEV,"%s header too big",msg ? msg : "");
				free(head);
				return(NULL);
			}
			strlcat(buffer, head, sizeof(buffer));
			strlcat(buffer, "\r\n\t", sizeof(buffer));
			strlcat(buffer,  tail, sizeof(buffer));
			header_size += strlen(head) + 3;
			free(head);
		}
      		header_size += cur_len;
      		if (header_size + 2 > MAX_HDR_SIZE) { /* 2 for CR LF */
			syslog(LLEV,"%s header too big", msg ? msg : "");
			return(NULL);
		}
      		if (cur_len != 2 && WHITE (buffer[line_len])) {
	  	/* It's a continuation line.  Adjust the length of the line
             	   and continue reading.  Note that we just keep the leading
                   whitespace. */

	  		line_len += cur_len;
		} else {
	  	/* It's not a continuation line.  Move the previous line to
             	   the linked list.  TODO: Use fastheap. */

			p = (struct hdr_line *)xmalloc (sizeof (struct hdr_line));
		/* Strip *last* CRLF, keep intact "continuation" ones inside */
	  		p->line = xstrndup (buffer, line_len - 2);
	  		p->next = NULL;
	  		p->length = line_len - 2;

	  		*add = p; add = &p->next;

	  	/* If the new line is empty we're finished. */

	  		if (cur_len == 2)  
	    			return first;

		/* Special case for some protocols: a dot occured in header means
		   the header itself is incomplete, but it does not indicate a
		   protocol error (see nntp, pop3 and alikes). We are adding the
		   dot itself to let the calling code know that something happened. */

	  		if ((cur_len == 1) && (buffer[line_len] == '.')) {
				p = (struct hdr_line *)xmalloc (sizeof (struct hdr_line));
	  			p->line = xstrndup (".", 1);
	  			p->next = NULL;
	  			p->length = line_len;
				*add = p;
	    			return first;
			}

	  /* Otherwise, move the line to the beginning of the buffer
             to make room for appending continuation lines. */

	  		memmove (buffer, buffer + line_len, cur_len);
	  		line_len = cur_len;
		}
    	}
/*
 *  Unreached
 */
  return first;
}

/* Check if the header was "gracefully truncated" by email transfer 
 * protocols like pop3, nntp and alikes.
 */

int
header_is_truncated (const struct hdr_line *list)
{
  	while (list->next)
    		list = list->next;
	return ((list->length == 1) && (*(list->line) == '.'));
}

/* Fake a header.line */

void 
header_add_line (struct hdr_line **list, const char *s)
{
  	struct hdr_line *p;

  	p = (struct hdr_line *)xmalloc (sizeof (struct hdr_line));
  	p->length = strlen (s);
  	p->line = xstrndup (s, p->length);
  	p->next = NULL;

  	/* This is slow but that shouldn't matter here. */

  	while (*list != NULL)
    		list = &(*list)->next;
  	*list = p;
}

void
header_insert_line (struct hdr_line **list, const char *s)
{
  	struct hdr_line *p;

  	p = (struct hdr_line *)xmalloc (sizeof (struct hdr_line));
  	p->length = strlen (s);
  	p->line = xstrndup (s, p->length);
	p->next = *list;
	*list = p;
}


/* Write the lines from the linked list pointed to by P to the file
   F. */

void 
header_write (EMO_FILE *f, const struct hdr_line *p)
{
  	while (p && !emo_error(f)) {
      		emo_write (f, p->line, p->length);
      		emo_write (f, "\r\n", 2);
      		p = p->next;
    	}
}

/* Free memory allocated by headers
 */

void header_free(struct hdr_line **headers)
{
	struct hdr_line *h;
	h = *headers;
	while (h) {
		struct hdr_line *h1;
		h1 = h;
		h = h->next;
		free(h1->line);
		free(h1);
	}
	*headers = NULL;
}

void header_drop(struct hdr_line **headers, struct hdr_line *drop)
{
	struct hdr_line *h;
		
	if ((h = drop->next)) {
	/* Not the last element, substitute with next in chain */
		free(drop->line);
		drop->line = h->line;
		drop->length = h->length;
		drop->next = h->next;
		free(h);
	} else {
	/* Last element, unlink */
		if (*headers == drop) {
	/* Last and only 	*/
			header_free(headers);
			return;
		}
		for (h = *headers ; h ; h = h->next) {
			if (h->next == drop)
				break;
			if (h->next == NULL) {
				syslog(LLEV,"fwtksyserr: internal logic error, attempt to remove non-existant header");
				exit(1);
			}
		}
		h->next = NULL;
		free(drop->line);
		free(drop);
	}
		
}


