/*
 * Copyright (c) 2009 Security Technology Research
 * All rights reserved.
 */
/*
 * Copyright (c) 1997-2007, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/* HTTP protocol support fucntions */

/* Copyright 1997-2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <inttypes.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"
#include "auth.h"
#include "ctype.h"
#include "headers.h"
#include "http.h"
#include "icap.h"

/* Statistics */

/* =========================== PRELIMINARY STUFF =========================== */

/* Maximum total size of header. */

#define MAX_HDR_SIZE		(128 * 1024)

/* Flag bits for method_flags in struct request_state. */

#define METHOD_REQUEST_BODY	0x0001 /* Request has a body */
#define METHOD_RESPONSE_BODY	0x0002 /* Response may have a body */
#define METHOD_CONNECT		0x0004 /* CONNECT, handled specially */


/* This buffer is used for various tasks, therefore we have to be
   cautious about not overwriting it in functions which take pointers
   which may point into this buffer.  Using multiple buffers would
   enlarge the working set. */

static char buffer[65536];

/* ============================== CHARACTERS =============================== */

/* Character classes as defined by RFC 1945. */

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

/* Also referenced by cookies.c. */

const unsigned char http_char_type[UCHAR_MAX+1] =
{
#include "http-ctype.h"
};

/* =========================== INPUT AND OUTPUT ============================ */


static void 
update_status(EMI_FILE *src, EMO_FILE *dst, direction d)
{
	switch (d) {
	case NONE:
		return;
	case IN:
		proxy_stats.inbytes = emi_amount(src, 0);
		break;
	case OUT:
		proxy_stats.outbytes = emo_amount(dst, 0);
	}
	proxy_update_status();
}

/* Copy up to SIZE bytes from file SRC to file DST. */

void http_copy (dst, chunked_out, src, chunked_in, size, d)
EMO_FILE	*dst;
int		chunked_out;
EMI_FILE	*src;
int		chunked_in;
long long	size;
direction	d;
{
  	size_t 	n;
  	char 	chunksize[16];
  	int 	firstchunk = 1;

  	if (size == 0)
    		return;
  	if (size == -1)
    		size = MAX_CONTENT_LENGTH;
  	if (!chunked_in) {
    		while (size > 0) {
        		n = (size < (long)sizeof (buffer)) ? 
					(size_t)size : sizeof (buffer);
        		n = emi_read (src, buffer, n);
        		if (n <= 0)
	  			break;
			if (chunked_out) {
	  			if (firstchunk) 
	    				firstchunk = 0;
	  			else 
            				emo_write (dst, "\r\n", 2);
	  			snprintf(chunksize,14,"%x\r\n",(unsigned int) n);
          			emo_write (dst, chunksize, strlen(chunksize));
	  			if (emo_error(dst))
	    				return;
			}
       			emo_write (dst, buffer, n);
			if (emo_error(dst))
	  			return;
        		size -= n;
			update_status(src,dst,d);
      		}
    		if (chunked_out) 
      			emo_write (dst, "\r\n0\r\n\r\n", 7);
    		if (emi_error (src) || emo_error(dst))
      			return;
  	} else {
    		unsigned long int chunk_left = 0;
    		int i;
    		int c;
    		while (size > 0) {
			if(!chunk_left) {
	  			for (i = 0; (i < 14) && (c != CR) ; i++) {
	    				c = emi_getc(src);
	    				chunksize[i] = c;
	  			}
	  			if ((i >= 14) || ((c = emi_getc(src)) != LF)) {
	    				syslog(LLEV, "invalid chunk in data stream");
	    				return;
	  			}
	  			chunksize[i] = '\0';
	  			chunk_left = strtoul(chunksize, NULL, 16);
	  			if (!chunk_left && chunked_out) {
            				emo_write (dst, "0\r\n\r\n", 5);
	    				return;
	  			}
			}
        		n = (size < (long)sizeof (buffer)) ? 
					(size_t)size : sizeof (buffer);
			n = chunk_left < n ? chunk_left : n;
        		n = emi_read (src, buffer, n);
        		if (n <= 0)
	  			break;
			if (chunked_out) {
	  			snprintf(chunksize,14,"%x\r\n",(unsigned int) n);
          			emo_write (dst, chunksize, strlen(chunksize));
			}			
        		emo_write (dst, buffer, n);
			if (chunked_out)
          			emo_write (dst, "\r\n", 2);
			if (emo_error(dst))
	  			return;
        		size -= n;
			chunk_left -= n;
			if (!chunk_left)
          			n = emi_read (src, buffer, 2);
			update_status(src,dst,d);
      		}
    		if (chunked_in && size == 0) {
          		n = emi_read (src, buffer, 5);
	  		if (n != 5 || strncmp(buffer, "0\r\n\r\n",5))
	    			syslog(LLEV,"warning: unterminated chunked stream ");
    		}
  	}
  	update_status(src,dst,d);
}

/* Save up to SIZE bytes from file SRC to a temporary file. */

char* http_save (char *template, EMI_FILE *src, int chunked, long long size)
{
  	char 		*qf;
  	int		tfd;
  	EMO_FILE 	*tqf;

  	qf = xstrdup(template);

  	if ((tfd = mkstemp(qf)) == -1) {
      		syslog(LLEV,"fwtksyserr: cannot create temporary file, %s", 
			strerror(errno));
      		exit(1);
    	}

  	tqf = emo_fdopen(tfd, 4096);
  	size = (size == -1) ? MAX_CONTENT_LENGTH : size;
  	http_copy (tqf, 0, src, chunked, size, NONE);
	if (emo_error(tqf)) {
      		syslog(LLEV,"fwtksyserr: cannot write temporary file");
		exit(1);
	}
	if (emi_error(src)) {
      		syslog(LLEV,"read error");
  		emo_close(tqf);
		unlink(qf);
		return(NULL);
	}
  	emo_close(tqf);
  	return(qf);
}

void http_load (char *qfile, EMO_FILE *dst, long long size)
{
  	size_t 	n;
  	int	tfd;

  	if ((tfd = open(qfile, O_RDONLY)) == -1) {
      		syslog(LLEV,"fwtksyserr: cannot open temporary file, %s", strerror(errno));
      		exit(1);
    	}

  	while (size > 0 && (!emo_error(dst))) {
      		n = (size < (long)sizeof (buffer)) ? 
				(size_t)size : sizeof (buffer);
      		n = read (tfd, buffer, n);
      		if (n <= 0)
			break;
      		emo_write (dst, buffer, n);
      		size -= n;
    	}	
}


/* Write the lines from the linked list pointed to by P to the file
   F. */

void http_write_header (EMO_FILE *f, const struct hdr_line *p)
{
  	while (p && !emo_error(f)) {
      		emo_write (f, p->line, p->length);
      		emo_write (f, "\r\n", 2);
      		p = p->next;
    	}
}

