/*
 * Copyright 2009 Security Technology Research
 * All rights reserved. 
 */

/* Internet Content Adaptation Protocol (RFC3507) support functions
 */

#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "headers.h"
#include "http.h"
#include "icap.h"
#include "url.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"
#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))
#define WHITE(c)        HTTP_CHAR_TYPE (c, CTYPE_WHITE)

#ifndef MIN
# define MIN(a, b) (((a) > (b)) ? (b) : (a))
#endif

extern const unsigned char http_char_type[UCHAR_MAX+1];

static int handle_icap_version (const char *line)
{
	return (strncmp("ICAP/1.0",line, 8) ? -1 : 8);
}

static int http_hdr_size(struct hdr_line *hdr)
{
	int size = 0;
	struct hdr_line *p;
	for (p = hdr; p ; p = p->next)
		size += p->length + 2;
	return (size);
}
static int icap_return_code(const char *line)
{
  	int j;

  	j = handle_icap_version (line);
  	if (j < 0 || !WHITE (line[j]))
    		return 0;
  	while (WHITE (line[j]))
    		++j;
  	return (atoi(line + j));
}

static int
parse_encap(const char *line, const char *what)
{
	char *p;
	return ((p = strstr(line, what)) ? atoi(p + strlen(what)) : -1);
}

int
icap_is_response(struct hdr_line *icap_hdr)
{
	char	*encap_line;
	
	encap_line = header_field_value(header_field_by_name(icap_hdr,"Encapsulated:"), NULL);
	if (!encap_line)
		return(0);
	return (parse_encap(encap_line, "res-hdr=") != -1);
}

int 
icap_open_connection(char *icap_url, EMI_FILE **in, EMO_FILE **out)
{
	int icap_fd;
	char icap_host[1024];
	char errmsg[MAX_STR];
	struct url u;
	if (url_parse(&u, (octet *) icap_url, strlen(icap_url), 0) || 
	    u.host_length > MAX_STR - 6) {
		syslog(LLEV,"fwtkcfgerr: bad icap url: %.128s", icap_url);
		return (-1);
	}
	strlcpy(icap_host, icap_url + u.host_start,
	        MIN(sizeof(icap_host), u.host_length +1));
	if ((icap_fd = conn_server(icap_host, u.port, 0, errmsg)) < 0) {
		syslog(LLEV,"fwtksyserr: cannot connect to %.128s:%u (%s)", 
			    icap_host, u.port, errmsg);
		return(-1);
	}
	if ((*in = emi_fdopen(icap_fd, 4096)) &&
	    (*out = emo_fdopen(icap_fd, 4096)))
		return (0);	
	else
		return (-1);

}

/*
 * Perfrom an ICAP request.
 * There is a reason behind keeping pre-modification headers separate
 * and flushing the whole output to EMO_FILE stream headers and body
 * altogether. We just do not want to have post-icap header callback
 * procedure and we want to avoid creation of unnecessary temporary files,
 * as we tend to do some builtin headers processing and we usually do
 * that before ICAP. For REQMOD EMO_FILE should point to the server
 * connection, as we send the request immediately, for RESPMOD we talk
 * to client providing modified response.
 */

int 
icap_request(EMI_FILE *icap_in, EMO_FILE *icap_out, const char *icap_url, 
	     icap_req icap_request, 
	     struct hdr_line *req_hdr, struct hdr_line *res_hdr,
	     EMI_FILE *body, int chunked, int bodysize, 
	     char **cachefile, int previewsize, struct hdr_line **icap_hdr)
{
	char	buffer[MAX_STR * 4];
	char	icap_host[MAXPATHLEN];
	struct  url u;
	int rc;

#define CHECK_ERROR if(emo_error(icap_out) || emi_error(icap_in)) return (502)

	if (url_parse(&u, (octet *) icap_url, strlen(icap_url), 0) || 
	    u.host_length > MAX_STR - 6) {
		syslog(LLEV,"fwtkcfgerr: bad icap url: %.128s", icap_url);
		return 400;
	}
	strlcpy(icap_host, icap_url + u.host_start,
	        MIN(sizeof(icap_host), u.host_length + 1));

	/* We do not preview chunked content. Most http servers provide things
	 * we are going to inspect as non-chunked, so let's just accept this
 	 * minor performance drawback. It is just a few extra lines to handle 
	 * that properly, but it is quite ugly when it comes to memory allocation
 	 * and stuff.
	 */

	switch(icap_request) {
	case OPTIONS:
	 	emo_printf(icap_out, "OPTIONS %.256s ICAP/1.0\r\n", icap_url);
		emo_printf(icap_out, "Host: %.256s\r\n", icap_host);
		emo_printf(icap_out, "X-Client-IP: %.256s\r\n", proxy_stats.riaddr);
		/* XXXX */
		break;
	case REQMOD:
	 	emo_printf(icap_out, "REQMOD %.256s ICAP/1.0\r\n", icap_url);
		emo_printf(icap_out, "Host: %.256s\r\n", icap_host);
		emo_printf(icap_out, "X-Client-IP: %.256s\r\n", proxy_stats.riaddr);
		emo_printf(icap_out, "Allow: 204\r\n");
		if (body) 
			emo_printf(icap_out,
			 	 "Encapsulated: req-hdr=0, req-body=%u\r\n",
				 http_hdr_size(req_hdr) + 2);
		else
			emo_printf(icap_out,
			 	 "Encapsulated: req-hdr=0, null-body=%u\r\n",
				 http_hdr_size(req_hdr) + 2);
		if (body && previewsize && !chunked)
			emo_printf(icap_out, "Preview: %u\r\n", previewsize);			
		break;
	case RESPMOD:
	 	emo_printf(icap_out, "RESPMOD %.256s ICAP/1.0\r\n", icap_url);
		emo_printf(icap_out, "Host: %.256s\r\n", icap_host);
		emo_printf(icap_out, "X-Client-IP: %.256s\r\n", proxy_stats.riaddr);
		emo_printf(icap_out, "Allow: 204\r\n");
		emo_printf(icap_out, "Encapsulated: ");
		if (req_hdr)
			emo_printf(icap_out, "req-hdr=0, ");
		if (res_hdr)
			emo_printf(icap_out, "res-hdr=%u, ", 
					     http_hdr_size(req_hdr) + 2);
		if (body)
			emo_printf(icap_out, "res-body=%u\r\n", 
					     http_hdr_size(req_hdr) +
					     http_hdr_size(res_hdr) + 4  );
		else 
			emo_printf(icap_out, "null-body=%u\r\n ", 
					     http_hdr_size(req_hdr) +
					     http_hdr_size(res_hdr) +4);
		if (body && previewsize && !chunked)
			emo_printf(icap_out, "Preview: %u\r\n", previewsize);			
	}
	emo_write(icap_out, "\r\n", 2);
	if (req_hdr) {
		header_write(icap_out, req_hdr);
		emo_write(icap_out, "\r\n", 2);
	}
	if (icap_request == RESPMOD) 
		header_write(icap_out, res_hdr);

	emo_flush(icap_out);
	CHECK_ERROR;

	if (body) {
		EMI_FILE *qf_in = NULL;
		if (icap_request == RESPMOD) 
			emo_write(icap_out, "\r\n", 2);
		if (previewsize && !chunked) {
			char *lbuffer;
			int n;
			int preview = MIN(previewsize, (bodysize == -1) ? previewsize : bodysize);

			if (!*cachefile) {
				n = emi_fill(body, preview);
				if (n == -1)
					return(-1);
			} else if (!(qf_in = emi_open(*cachefile, previewsize + 1024))) {
				syslog(LLEV,"fwtksyserr: cannot open cache file %s", 
					     *cachefile);
				return(502);
			}

		/*	We *MUST* have body size at this point for REQMOD, as we cannot say
			where it ends while connection is not closed. But there is no need
			to handle that any special way; let's just wait forever (until network
			timeout) if there is not enough data as HTTP servers do.
		 */

			lbuffer = xmalloc(previewsize);
			n = emi_peek(qf_in ? qf_in : body, lbuffer, preview);
			if (n == -1)
				return(qf_in ? 502 : -1);
			if (n) {
				emo_printf(icap_out, "%x\r\n", n);
				emo_write(icap_out, lbuffer, n);
				emo_printf(icap_out, "\r\n");
			}
			if (n == previewsize) {
				emo_printf(icap_out, "0\r\n\r\n");
				emo_flush(icap_out);
				CHECK_ERROR;

				emi_gets(icap_in, buffer, sizeof(buffer) - 2); 
				CHECK_ERROR;

				switch (rc = icap_return_code(buffer)) {
				case 100:	break;
				case 200:	goto got_200;
				case 204:	return(rc);
				default:	syslog(LLEV, "icap server returned \"%.128s\" on preview", buffer);
						return(rc);
				}
				/* 100 Continue
				 * that means we should cache the body for future use
			 	 */
				if (!*cachefile) {
					*cachefile = http_save ("/tmp/sqgw.XXXXXXXXX",
								body, chunked,
								bodysize == -1 ?
								1024 * 1024 * 1024 : 
								bodysize);
					if (emi_error(body))
						return(502);
					CHECK_ERROR;
					DEBUG_ASSERT(!qf_in);
					if (!(qf_in = emi_open(*cachefile, previewsize + 1024))) {
						syslog(LLEV,"fwtksyserr: cannot open cache file %s", 
					     			*cachefile);
						return(502);
					}
				}	
				n = emi_read(qf_in, lbuffer, previewsize);
				DEBUG_ASSERT (n == previewsize); 
				DEBUG_ASSERT ((bodysize == -1) || (bodysize >= n)); 
				http_copy(icap_out, 1, qf_in , 0, 
					  bodysize == -1 ? 1024 * 1024 * 1024 : bodysize - n, NONE);
				if (emi_error(qf_in))
					return(502);
				emi_close(qf_in);
				CHECK_ERROR;
			} else 
				emo_printf(icap_out, "0; ieof\r\n\r\n");
		} else {
			if (!*cachefile) 
				*cachefile = http_save ("/tmp/sqgw.XXXXXXXXX",
							body, chunked,
							bodysize == -1 ?
							1024 * 1024 * 1024 : 
							bodysize);
			DEBUG_ASSERT(!qf_in);
			if ((!*cachefile) || !(qf_in = emi_open(*cachefile, previewsize + 1024))) {
				syslog(LLEV,"fwtksyserr: cannot open cache file %s", 
			     			*cachefile ? *cachefile : "(not saved)");
				return(502);
			}
			http_copy(icap_out, 1, qf_in, 0, 
				  bodysize == -1 ? 1024 * 1024 * 1024 : 
						   bodysize, NONE);
			emi_close(qf_in);
			if (emi_error(qf_in))
				return(502);
			CHECK_ERROR;
		}
	}
	emo_flush(icap_out);
	CHECK_ERROR;

	do
		emi_gets(icap_in, buffer, sizeof(buffer) - 2); 
	while(!*buffer && !emi_error(icap_in));
	CHECK_ERROR;

	if ((rc = icap_return_code(buffer)) != 200) { 
		if (rc != 204)
			syslog(LLEV, "icap server returned \"%.128s\"", buffer);
		return(rc);
	}

got_200:*icap_hdr = header_read(icap_in, "icap");
	CHECK_ERROR;
	return(rc);
}
