/*
 * Copyright (c) 1999-2009, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/* connect.c */

/* Copyright 1999-2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.

   1999-03-17 Initial version
   2000-05-31 copy_bidir() has an additional argument */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <limits.h>
#include <ctype.h>
#ifdef  OpenBSD
#include        <ssl/x509.h>
#include        <ssl/x509v3.h>
#include        <ssl/err.h>
#include        <ssl/ssl.h>
#else
#include        <openssl/x509.h>
#include        <openssl/x509v3.h>
#include        <openssl/err.h>
#include        <openssl/ssl.h>
#include        <openssl/rand.h>
#include        <openssl/bn.h>
#endif

#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "sslfunc.h"
#include "emio.h"
#include "url.h"
#include "squid-gw.h"
#include "headers.h"
#include "http.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"
#include "icap.h"

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))
extern const unsigned char http_char_type[UCHAR_MAX+1];

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

#define HOST(c) (isascii(c) && (isalnum(c) || (c == '.') || (c == '-')))

extern struct hdr_line *req_hdr;
struct hdr_line *old_hdr;
extern struct hdr_line *res_hdr;
extern struct request_state req_st;
extern struct response_state res_st;

int parse_connect (char *hn, size_t hnsize, int *port, const char *src)
{
  	size_t i = 0;
  	char *end;
  	long n;

  	while (HOST (src[i]))
    		++i;
  	if (i == 0 || i >= hnsize)
    		return -1;
  	memcpy (hn, src, i); hn[i] = 0;
  	src += i;
  	if (*src++ != ':')
    		return -1;			/* The port number is mandatory */
  	n = strtol (src, &end, 10);
  	if (n < 0 || n > 65535 || end == src)
    		return -1;
  	*port = (int)n;
  	src = end;
  	return *src == ' ' ? 0 : -1;
}

/* Return address list corresponding to given hostname;
 * NULL if dns is not available
 */

char **connect_addr_list (const char *hostname)
{
  	static octet 	addr[4];
  	octet 		mask[4];
  	struct hostent 	*h;

	if (cf_http.forward_connect && proxy_nodns) 
		return (NULL);

  	if (ipaddr_parsez ((octet*) hostname, addr, mask, 0) == 0) {
      		static char *vec[2];
      		vec[0] = (char *)addr;
      		vec[1] = NULL;
      		return vec;
    	}
  	h = gethostbyname ((char *)hostname);
  	if (h == NULL) {
		if (!cf_http.forward_connect)
      			syslog (LLEV, "%.512s host name lookup failed", hostname);
      		return NULL;
    	}
  	if (h->h_addrtype != AF_INET || h->h_length != 4) {
      		syslog (LLEV, "%.512s: invalid h_addrtype or h_length",
	      		hostname);
      		return NULL;
    	}
  	return h->h_addr_list;
}

static int handle_status_line (const char *line)
{
  	int 	i, j = 0;
  	int 	code;

  	/* We do not actually handle version here. Should we? */

  	if (strlen(line) < 8 || strncmp (line, "HTTP/", 5) != 0 || !DIGIT (line[5]))
    		return -1;

  	while (!WHITE (line[j]))
    		++j;
  	while (WHITE (line[j]))
    		++j;
  	code = 0;
  	for (i = 0; i < 3; ++i) {
      		if (!DIGIT (line[j+i]))
			return -1;
      		code = code * 10 + line[j+i] - '0';
    	}
  	/* Some web servers omit the (mandatory) space after the status
     	code. */
  	if (!WHITE (line[j+3]) && line[j+3] != 0)
    		return -1;
  	return code;
}

/* All the following code is just ugly workaround for Mozilla's way 
 * they [mis]designed to fix CVE-2009-1836.
 * Hereby i send Mighty Ray of Bloody Diarrhea and Rectal Cracks to
 * Mozilla developers responsible for that "solution". 
 * Burn in Hell, fucking freaks!
 * See details here: https://bugzilla.mozilla.org/show_bug.cgi?id=479880
 * Actually, the code was good for some reuse anyways.
 */

/* Forge an SSL connection so we can provide error page
 */

static void
SSL_error_prep(const char *srvname, const char* http_version) 
{
	emo_printf (c_out, "HTTP/%s 200 Connection established\r\n\r\n", http_version);
  	if (emo_flush (c_out) != 0) {
      		syslog (LLEV, "emo_flush: %s", strerror(errno));
      		exit (1);
    	}
  	emio_sslaccept(c_in, c_out,  
		       proxy_stats.rladdr, proxy_stats.riaddr, "none",
		       proxy_SSL_makeup_error_cert(srvname), NULL);

	/* Not actually a fwtksyserr and we cannot present http_error
	   here as we are expected to switch to encryption connection
	   already. Just fail and fall out instead.
	 */

	if (!c_in->sslio) {
		syslog(LLEV, "Cannot negotiate client SSL conenction");	
		exit(1);
	}
}

/* End of infinite uglyness.
 */


void http_connect2 (const octet *ipaddr, int port, const char *http_version,
		    int timeout)
{
  	int r, fd;
	char buffer[1024];

  	if (cf_http.forward_connect) {
		if ((fd = conn_server(cf_http.server, 
				cf_http.port, 0, buffer)) < 0) 
			http_error(503, "Upstream proxy error: %s", buffer);
	} else { /* Ugly, but we need to be sure ip address 
		    is exactly the same */
		if ((fd = conn_server(inet_ntoa(*((struct in_addr*)ipaddr)), 
				port, 0, buffer)) < 0) 
			http_error(503, "Cannot connect: %s", buffer);
    	}

  	s_in = emi_fdopen (fd, S_IN_BUFSIZE);
  	s_out = emo_fdopen (fd, 4096);

  	if (cf_http.forward_connect) {
		emo_printf (s_out, "CONNECT %s:%u HTTP/%s\r\n\r\n", proxy_stats.dst, port,
				http_version);
		if (emo_flush (s_out) != 0) {
			syslog (LLEV, "emo_flush: %s", strerror(errno));
			exit (1);
		}
		emi_gets(s_in, buffer, sizeof(buffer));
		if (handle_status_line(buffer) != 200) 
			http_error(503, "Upstream proxy error: %s", buffer);
		/* trailing CRLF */
		emi_gets(s_in, buffer, sizeof(buffer));
	}
  	switch (cf_http.connect) {
  	case CONNECT_TUNNEL:
  	case CONNECT_INSPECT:
		emo_printf (c_out, "HTTP/%s 200 Connection established\r\n\r\n",
			    http_version);
		res_st.code = 200;
		if (emo_flush (c_out) != 0) {
			syslog (LLEV, "emo_flush: %s", strerror(errno));
			exit (1);
		}
  		s_in = emi_fdopen (fd, S_IN_BUFSIZE);
		r = copy_bidir_emi (s_in, fd, c_in, -1, timeout, CB_EOF1 | CB_EOF2);
		if (r > 0)
			syslog (LLEV, "timeout");
		if (r < 0)
			exit (1);
		return;
 	 case CONNECT_MITM:
		/* Override ciphers and options, as we may get it from the class policy */

		if (cf_http.ciphers)
			proxy_SSLciphers = cf_http.ciphers;
		else
			proxy_SSLciphers = "HIGH:-MD5";

		if (!SSL_CTX_set_cipher_list(proxy_SSLCTX, proxy_SSLciphers)) {
			syslog(LLEV,"fwtksyserr: error selecting cipher list %s: %.128s", 
				proxy_SSLciphers, ERR_error_string(ERR_get_error(), NULL));
				exit(1);
			}

		if (!cf_http.allow_sslv2.v)
			SSL_CTX_set_options(proxy_SSLCTX, SSL_OP_NO_SSLv2);
		else
			SSL_CTX_clear_options(proxy_SSLCTX, SSL_OP_NO_SSLv2);

		/* We do not pass verify parameter as we handle certificate
		   validation separately */
 		emio_sslconnect(s_in, s_out, proxy_stats.dst, "none", cf_http.client_cert, cf_http.client_key);
		if (!s_in->sslio)	
			http_error(503, "Cannot negotiate SSL connection: %s", buffer);
		DEBUG_ASSERT(sizeof(buffer) > 512);
		r = proxy_SSLcheck_cert(s_out->sslio,proxy_stats.dst,
				strcmp(cf_http.server_vrfy,"strict") ? 
				"basic" : "checkaddr", buffer);
		if (PROXYSSL_CHECK_VRFY(cf_http.server_vrfy,r)) {
			SSL_error_prep(proxy_stats.dst,http_version);
			http_error(503,"Server certificate error: %s",
				   buffer);
					
		}
  		emo_printf (c_out, "HTTP/%s 200 Connection established\r\n\r\n",
	      		http_version);
  		if (emo_flush (c_out) != 0) {
      			syslog (LLEV, "emo_flush: %s", strerror(errno));
      			exit (1);
    		}
  		emio_sslaccept(c_in, c_out,  
			       proxy_stats.rladdr, proxy_stats.riaddr, "none",
			       proxy_SSLspoof_server_cert(s_in->sslio), NULL);

		/* Not actually a fwtksyserr and we cannot present http_error
		   here as we are expected to switch to encryption connection
		   already. Just fail and fall out instead.
		 */

		if (!c_in->sslio) {
			syslog(LLEV, "Cannot negotiate client SSL conenction");	
			exit(1);
		}

		/* TODO: handle persistent connections.
	 	 * while (1) {
	 	 */


   		http_request(c_in, s_out);
		http_response (s_in, c_out);

		/* } 
	 	*/
  	}

}
